﻿namespace TabsStudioSorter
{
    class Comparer
    {
        public Comparer(Options options_, AddinsCommon.IsPreviewTabProperty isPreviewTabProperty_,
            System.Collections.Generic.Dictionary<System.Windows.Controls.TabItem, FixedTabInfo> fixedInfo_,
            Marker marker_)
        {
            options = options_;
            isPreviewTabProperty = isPreviewTabProperty_;
            fixedInfo = fixedInfo_;
            marker = marker_;
        }

        public int CompareTabs(System.Windows.Controls.TabItem lhs, System.Windows.Controls.TabItem rhs)
        {
            if (isPreviewTabProperty.IsPreviewTab(lhs))
                return 1;
            if (isPreviewTabProperty.IsPreviewTab(rhs))
                return -1;

            FixedTabInfo lhsInfo = fixedInfo[lhs];
            FixedTabInfo rhsInfo = fixedInfo[rhs];
            foreach (SortingRule rule in options.rules)
            {
                int result = CompareByRule(lhs, rhs, lhsInfo, rhsInfo, rule);
                if (result != 0)
                    return result;
            }
            return 0;
        }

        private int CompareByRule(System.Windows.Controls.TabItem lhs, System.Windows.Controls.TabItem rhs, 
            FixedTabInfo lhsInfo, FixedTabInfo rhsInfo,
            SortingRule rule)
        {
            int result = 0;
            switch (rule.type)
            {
            case SortingRuleType.MarkerHighlightIndex:
                uint lhsIndex = marker.TabHighlightIndex(lhs);
                uint rhsIndex = marker.TabHighlightIndex(rhs);
                if (lhsIndex == 0)
                    lhsIndex = uint.MaxValue;
                if (rhsIndex == 0)
                    rhsIndex = uint.MaxValue;
                if (lhsIndex != rhsIndex)
                {
                    return ApplyOrderRule(lhsIndex < rhsIndex ? -1 : 1, rule);
                }
            	break;
            case SortingRuleType.Project:
                result = CompareMatchedStrings(lhsInfo.project, rhsInfo.project, rule);
                if (result != 0)
                    return ApplyOrderRule(result, rule);
                break;
            case SortingRuleType.Extension:
                result = CompareMatchedStrings(lhsInfo.extension, rhsInfo.extension, rule);
                if (result != 0)
                    return ApplyOrderRule(result, rule);
                break;
            case SortingRuleType.TabName:
                result = CompareMatchedStrings(GetTabNameForSorting(lhs, false), GetTabNameForSorting(rhs, false), rule);
                if (result != 0)
                    return ApplyOrderRule(result, rule);
                break;
            case SortingRuleType.TabNameWithPath:
                result = CompareMatchedStrings(GetTabNameForSorting(lhs, true), GetTabNameForSorting(rhs, true), rule);
                if (result != 0)
                    return ApplyOrderRule(result, rule);
                break;
            case SortingRuleType.IsDocument:
                if (lhsInfo.isDocument != rhsInfo.isDocument)
                    return ApplyOrderRule(lhsInfo.isDocument ? -1 : 1, rule);
                break;
            case SortingRuleType.Path:
                result = CompareMatchedStrings(lhsInfo.paths, rhsInfo.paths, rule);
                if (result != 0)
                    return ApplyOrderRule(result, rule);
                break;
            } 
            return 0;
        }

        private int CompareMatchedStrings(string lhs, string rhs, SortingRule rule)
        {
            string lhsMatch = ExtractM(lhs, rule);
            string rhsMatch = ExtractM(rhs, rule);

            if (lhsMatch == null && rhsMatch == null)
                return 0;

            if (lhsMatch != null && rhsMatch != null)
                return System.String.Compare(lhsMatch, rhsMatch, true);

            return lhsMatch != null ? -1 : 1;
        }

        private string ExtractM(string input, SortingRule rule)
        {
            if (input == null || rule.match == null || rule.match.Length == 0)
                return null;

            System.Text.RegularExpressions.Regex reg = new System.Text.RegularExpressions.Regex(rule.match);
            System.Text.RegularExpressions.Group m = reg.Match(input).Groups["M"];

            if (!m.Success)
                return null;

            return m.Value;
        }

        private int ApplyOrderRule(int result, SortingRule rule)
        {
            return rule.inverseOrder ? -result : result;
        }

        private string GetTabNameForSorting(System.Windows.Controls.TabItem tabItem, bool withPath)
        {
            string tabName = GetTabName(tabItem);
            if (withPath)
                return tabName;

            int pathEnd = tabName.LastIndexOf('/');
            return pathEnd >= 0 ? tabName.Substring(pathEnd + 1) : tabName;
        }

        private string GetTabName(System.Windows.Controls.TabItem tabItem)
        {
            System.Windows.Controls.DockPanel tabInternals = tabItem.Header as System.Windows.Controls.DockPanel;
            System.Windows.Controls.StackPanel tabNameGroup = tabInternals.Children[0] as System.Windows.Controls.StackPanel;
            System.Windows.Controls.Label tabName = tabNameGroup.Children[0] as System.Windows.Controls.Label;
            return tabName.Content as string;
        }

        private Options options;
        private AddinsCommon.IsPreviewTabProperty isPreviewTabProperty;
        private System.Collections.Generic.Dictionary<System.Windows.Controls.TabItem, FixedTabInfo> fixedInfo;
        private Marker marker;
    }
}
